# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: row_handler.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='row_handler.proto',
  package='bytegain.custom.model_data.row_handler_proto',
  syntax='proto2',
  serialized_pb=_b('\n\x11row_handler.proto\x12,bytegain.custom.model_data.row_handler_proto\"R\n\nRowHandler\x12\x44\n\x06\x62ucket\x18\x01 \x03(\x0b\x32\x34.bytegain.custom.model_data.row_handler_proto.Bucket\"\x9e\x03\n\x06\x42ucket\x12\x0c\n\x04name\x18\x01 \x01(\t\x12T\n\x0b\x62ucket_type\x18\x02 \x01(\x0e\x32?.bytegain.custom.model_data.row_handler_proto.Bucket.BucketType\x12J\n\textractor\x18\x03 \x01(\x0b\x32\x37.bytegain.custom.model_data.row_handler_proto.Extractor\x12\x10\n\x08has_null\x18\x04 \x01(\x08\x12\x14\n\x0c\x62ucket_count\x18\x05 \x01(\x05\x12\x15\n\rfloat_buckets\x18\x06 \x03(\x02\x12\x16\n\x0estring_buckets\x18\x07 \x03(\t\x12\x13\n\x0bint_buckets\x18\x08 \x03(\x05\x12\x17\n\x0f\x62ucket_fraction\x18\t \x03(\x02\x12\x12\n\nnone_value\x18\n \x01(\x02\"K\n\nBucketType\x12\n\n\x06\x44IRECT\x10\x00\x12\n\n\x06STRING\x10\x01\x12\n\n\x06\x42UCKET\x10\x02\x12\x0b\n\x07\x42OOLEAN\x10\x03\x12\x0c\n\x08\x43\x41TEGORY\x10\x04\"\x86\x04\n\tExtractor\x12]\n\x0e\x65xtractor_type\x18\x01 \x01(\x0e\x32\x45.bytegain.custom.model_data.row_handler_proto.Extractor.ExtractorType\x12\x12\n\nfield_name\x18\x02 \x01(\t\x12\n\n\x02mu\x18\x03 \x01(\x02\x12\r\n\x05sigma\x18\x04 \x01(\x02\x12\x14\n\x0c\x66ield_name_2\x18\x05 \x01(\t\x12\x0f\n\x07min_val\x18\x06 \x01(\x02\x12\x0f\n\x07max_val\x18\x07 \x01(\x02\x12\x13\n\x0bpercentiles\x18\x08 \x03(\x02\x12S\n\tdata_type\x18\t \x01(\x0e\x32@.bytegain.custom.model_data.row_handler_proto.Extractor.DataType\x12\x15\n\rnull_fraction\x18\n \x01(\x02\"|\n\rExtractorType\x12\n\n\x06\x44IRECT\x10\x00\x12\r\n\tAUTO_NORM\x10\x01\x12\x0c\n\x08NON_NULL\x10\x02\x12\x0c\n\x08LOG_NORM\x10\x03\x12\x13\n\x0fPERCENTILE_NORM\x10\x04\x12\t\n\x05RATIO\x10\x05\x12\x14\n\x10\x44\x45\x46\x41ULTED_DIRECT\x10\x06\"4\n\x08\x44\x61taType\x12\x07\n\x03INT\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\n\n\x06STRING\x10\x02\x12\x08\n\x04\x42OOL\x10\x03')
)



_BUCKET_BUCKETTYPE = _descriptor.EnumDescriptor(
  name='BucketType',
  full_name='bytegain.custom.model_data.row_handler_proto.Bucket.BucketType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DIRECT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BUCKET', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CATEGORY', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=491,
  serialized_end=566,
)
_sym_db.RegisterEnumDescriptor(_BUCKET_BUCKETTYPE)

_EXTRACTOR_EXTRACTORTYPE = _descriptor.EnumDescriptor(
  name='ExtractorType',
  full_name='bytegain.custom.model_data.row_handler_proto.Extractor.ExtractorType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DIRECT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AUTO_NORM', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NON_NULL', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG_NORM', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PERCENTILE_NORM', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RATIO', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEFAULTED_DIRECT', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=909,
  serialized_end=1033,
)
_sym_db.RegisterEnumDescriptor(_EXTRACTOR_EXTRACTORTYPE)

_EXTRACTOR_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='bytegain.custom.model_data.row_handler_proto.Extractor.DataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOL', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1035,
  serialized_end=1087,
)
_sym_db.RegisterEnumDescriptor(_EXTRACTOR_DATATYPE)


_ROWHANDLER = _descriptor.Descriptor(
  name='RowHandler',
  full_name='bytegain.custom.model_data.row_handler_proto.RowHandler',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bucket', full_name='bytegain.custom.model_data.row_handler_proto.RowHandler.bucket', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=67,
  serialized_end=149,
)


_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='bytegain.custom.model_data.row_handler_proto.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_type', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.bucket_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extractor', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.extractor', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_null', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.has_null', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_count', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.bucket_count', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='float_buckets', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.float_buckets', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_buckets', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.string_buckets', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int_buckets', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.int_buckets', index=7,
      number=8, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_fraction', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.bucket_fraction', index=8,
      number=9, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='none_value', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.none_value', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BUCKET_BUCKETTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=566,
)


_EXTRACTOR = _descriptor.Descriptor(
  name='Extractor',
  full_name='bytegain.custom.model_data.row_handler_proto.Extractor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='extractor_type', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.extractor_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='field_name', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.field_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mu', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.mu', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sigma', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.sigma', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='field_name_2', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.field_name_2', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_val', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.min_val', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_val', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.max_val', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='percentiles', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.percentiles', index=7,
      number=8, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.data_type', index=8,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='null_fraction', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.null_fraction', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _EXTRACTOR_EXTRACTORTYPE,
    _EXTRACTOR_DATATYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=569,
  serialized_end=1087,
)

_ROWHANDLER.fields_by_name['bucket'].message_type = _BUCKET
_BUCKET.fields_by_name['bucket_type'].enum_type = _BUCKET_BUCKETTYPE
_BUCKET.fields_by_name['extractor'].message_type = _EXTRACTOR
_BUCKET_BUCKETTYPE.containing_type = _BUCKET
_EXTRACTOR.fields_by_name['extractor_type'].enum_type = _EXTRACTOR_EXTRACTORTYPE
_EXTRACTOR.fields_by_name['data_type'].enum_type = _EXTRACTOR_DATATYPE
_EXTRACTOR_EXTRACTORTYPE.containing_type = _EXTRACTOR
_EXTRACTOR_DATATYPE.containing_type = _EXTRACTOR
DESCRIPTOR.message_types_by_name['RowHandler'] = _ROWHANDLER
DESCRIPTOR.message_types_by_name['Bucket'] = _BUCKET
DESCRIPTOR.message_types_by_name['Extractor'] = _EXTRACTOR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RowHandler = _reflection.GeneratedProtocolMessageType('RowHandler', (_message.Message,), dict(
  DESCRIPTOR = _ROWHANDLER,
  __module__ = 'row_handler_pb2'
  # @@protoc_insertion_point(class_scope:bytegain.custom.model_data.row_handler_proto.RowHandler)
  ))
_sym_db.RegisterMessage(RowHandler)

Bucket = _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), dict(
  DESCRIPTOR = _BUCKET,
  __module__ = 'row_handler_pb2'
  # @@protoc_insertion_point(class_scope:bytegain.custom.model_data.row_handler_proto.Bucket)
  ))
_sym_db.RegisterMessage(Bucket)

Extractor = _reflection.GeneratedProtocolMessageType('Extractor', (_message.Message,), dict(
  DESCRIPTOR = _EXTRACTOR,
  __module__ = 'row_handler_pb2'
  # @@protoc_insertion_point(class_scope:bytegain.custom.model_data.row_handler_proto.Extractor)
  ))
_sym_db.RegisterMessage(Extractor)


# @@protoc_insertion_point(module_scope)
